# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 23:36:07 2019

@author: ydima
"""

import time
import urllib

import pytest
import sqlalchemy as sa

from bcpandas import SqlCreds

from .utils import DockerDB

_db_name = "db_bcpandas"
_docker_startup = 15  # seconds to wait to give the container time to start
docker_db_obj = DockerDB(
    container_name="bcpandas-mssql-container", sa_sql_password="MyBigSQLPassword!!!"
)


def pytest_addoption(parser):
    parser.addoption("--mssql-docker-image", action="store")


@pytest.fixture(scope="session")
def docker_db(pytestconfig):
    # figure out which docker image
    docker_image = pytestconfig.getoption("--mssql-docker-image", default=None)
    if docker_image is not None:
        # overwrite default image of the DockerDB object
        docker_db_obj.mssql_image = docker_image

    # start it
    docker_db_obj.start()
    time.sleep(_docker_startup)
    print("successfully started DB in docker...")
    yield
    print("Stopping container")
    docker_db_obj.stop()
    print("Deleting container")
    docker_db_obj.remove()
    print("all done!")


@pytest.fixture(scope="session")
def database(docker_db):
    docker_db_obj.create_database(_db_name)


@pytest.fixture(scope="session")
def sql_creds():
    creds = SqlCreds(
        server=docker_db_obj.address,
        database=_db_name,
        username="sa",
        password=docker_db_obj.sa_sql_password,
    )
    return creds


@pytest.fixture(scope="session")
def pyodbc_creds(database):
    db_url = (
        "Driver={ODBC Driver 17 for SQL Server};"
        + f"Server={docker_db_obj.address};"
        + f"Database={_db_name};UID=sa;PWD={docker_db_obj.sa_sql_password};"
    )
    engine = sa.engine.create_engine(
        f"mssql+pyodbc:///?odbc_connect={urllib.parse.quote_plus(db_url)}"
    )
    return engine



import logging
import os
import warnings

import pandas as pd

from bcpandas.constants import (
    OUT,
    QUERY,
    QUERYOUT,
    SQL_TYPES,
    TABLE,
    VIEW,
    BCPandasValueError,
    read_data_settings,
)
from bcpandas.main import SqlCreds
from bcpandas.utils import bcp, get_temp_file

logger = logging.getLogger(__name__)


def read_sql(
    table_name: str,
    creds: SqlCreds,
    sql_type: str = "table",
    schema: str = "dbo",
    batch_size: int = None,
    debug: bool = False,
    delimiter: str = None,
    check_delim: bool = True,
    bcp_path: str = None,
) -> pd.DataFrame:
    """
    ** It is HIGHLY recommended to not use this method, and instead use the native pandas `read_sql*` methods. See README for details. **
    
    Reads a SQL table, view, or query into a pandas DataFrame.

    Parameters
    ----------
    table_name : str
        Name of SQL table or view, without the schema, or a query string
    creds : bcpandas.SqlCreds
        The credentials used in the SQL database.
    sql_type : {'table', 'view', 'query'}, default 'table'
        The type of SQL object that the parameter `table_name` is.
    schema : str, default 'dbo'
        The SQL schema of the table or view. If a query, will be ignored.
    batch_size : int, optional
        Rows will be read in batches of this size at a time. By default,
        all rows will be read at once.
    debug : bool, default False
        If True, will not delete the temporary CSV file, and will output its location.
    delimiter : str, optional
        One or more characters to use as a column delimiter in the temporary CSV file.
        If not supplied, the default used is specified in `constants.py` in the `read_data_settings` variable.
        **IMPORTANT** - the delimiter must not appear in the actual data in SQL or else it will fail.
    check_delim : bool, default True
        Whether to check the temporary CSV file for the presence of the delimiter in the data.
        See note below.
    bcp_path : str, default None
        The full path to the BCP utility, useful if it is not in the PATH environment variable
    
    Returns
    -------
    `pandas.DataFrame`

    Notes
    -----
    Will actually read the SQL table/view/query twice - first to get the names of the columns 
    (will only read first few rows), then all the rows using BCP.

    Also, the temporary CSV file will be read into memory twice, to check for the presence of
    the delimiter character in the data. This can cause it to take longer. If you are sure the
    delimiter isn't in the data, you can skip this check by passing `check_delim=False`
    """
    warnings.warn(
        "It is HIGHLY recommended to not use this method, and instead use the native pandas `read_sql*` methods. See README for details."
    )
    # check params
    assert sql_type in SQL_TYPES
    if batch_size == 0:
        raise BCPandasValueError("Param batch_size can't be 0")

    # set up objects
    if ";" in table_name:
        raise BCPandasValueError(
            "The SQL item cannot contain the ';' character, it interferes with getting the column names"
        )

    # read top 2 rows of query to get the columns
    logger.debug("Starting to read first 2 rows to get the column names")
    _from_clause = table_name if sql_type in (TABLE, VIEW) else f"({table_name})"

    _existing_data = pd.read_sql_query(
        sql=f"SELECT TOP 2 * FROM {_from_clause} as qry", con=creds.engine
    )

    if _existing_data.shape[0] > 0:
        cols = _existing_data.columns
        logger.debug("Successfully read the column names")
    else:
        return _existing_data

    file_path = get_temp_file()

    # set delimiter
    delim = delimiter if delimiter is not None else read_data_settings["delimiter"]
    try:
        bcp(
            sql_item=table_name,
            direction=QUERYOUT if sql_type == QUERY else OUT,
            flat_file=file_path,
            creds=creds,
            print_output=True,
            sql_type=sql_type,
            schema=schema,
            batch_size=batch_size,
            col_delimiter=delim,
            bcp_path=bcp_path,
        )
        logger.debug(f"Saved dataframe to temp CSV file at {file_path}")

        # check if delimiter is in the data more than it should be
        # there should be len(cols)-1 instances of the delimiter per row
        if check_delim:
            num_delims = len(cols) - 1
            with open(file_path, "r") as file:
                for line in file:
                    if line.count(delim) > num_delims:
                        raise BCPandasValueError(
                            f"The delimiter ({delim}) was found in the source data, cannot"
                            " import with the delimiter specified. Try specifiying a delimiter"
                            " that does not appear in the data."
                        )

        csv_kwargs = {}
        if len(delim) > 1:
            # pandas csv C engine only supports 1 character as delim
            csv_kwargs["engine"] = "python"

        return pd.read_csv(
            filepath_or_buffer=file_path,
            sep=delim,
            header=None,
            names=cols,
            index_col=False,
            **csv_kwargs,
        )
    finally:
        if not debug:
            logger.debug("Deleting temp CSV file")
            os.remove(file_path)
        else:
            logger.debug(
                f"`read_sql` DEBUG mode, not deleting the file. CSV file is at {file_path}."
            )



# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 23:20:19 2019

@author: ydima
"""

import os
import sys

import pandas as pd


class BCPandasException(Exception):
    pass


class BCPandasValueError(BCPandasException):
    pass


IS_WIN32 = sys.platform == "win32"

# BCP terms
IN = "in"
OUT = "out"
QUERYOUT = "queryout"
TABLE = "table"
VIEW = "view"
QUERY = "query"

DIRECTIONS = (IN, OUT, QUERYOUT)
SQL_TYPES = (TABLE, VIEW, QUERY)
IF_EXISTS_OPTIONS = ("append", "replace", "fail")


# Text settings
_DELIMITER_OPTIONS = (",", "|", "\t")
_QUOTECHAR_OPTIONS = ('"', "'", "`", "~")
NEWLINE = os.linesep

# settings for both BCP and pandas.read_csv for reading from SQL
# delimiter should be characters that NEVER appear in the source data in SQL, have to guess a good one
# note that in pandas.read_csv a delimiter longer than a single character is interpreted as a regex
# see https://docs.microsoft.com/en-us/sql/relational-databases/import-export/specify-field-and-row-terminators-sql-server
# and https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
read_data_settings = {"delimiter": "\\t", "newline": NEWLINE}

# BCP Format File terms
SQLCHAR = "SQLCHAR"
sql_collation = "SQL_Latin1_General_CP1_CI_AS"


error_msg = """Data contains all of the possible {typ} characters {opts}, 
cannot use BCP to import it. Replace one of the possible {typ} characters in
your data, or use another method besides bcpandas.

Further background:

https://docs.microsoft.com/en-us/sql/relational-databases/import-export/specify-field-and-row-terminators-sql-server#characters-supported-as-terminators
"""


def get_delimiter(df: pd.DataFrame) -> str:
    for delim in _DELIMITER_OPTIONS:
        if not df.applymap(lambda x: delim in x if isinstance(x, str) else False).any().any():
            return delim
    raise BCPandasValueError(error_msg.format(typ="delimiter", opts=_DELIMITER_OPTIONS))


def get_quotechar(df: pd.DataFrame) -> str:
    for qc in _QUOTECHAR_OPTIONS:
        if not df.applymap(lambda x: qc in x if isinstance(x, str) else False).any().any():
            return qc
    raise BCPandasValueError(error_msg.format(typ="quote", opts=_QUOTECHAR_OPTIONS))



# flake8: noqa F401
from subprocess import DEVNULL, run
import warnings

from .main import SqlCreds, to_sql
from .utils import bcp

name = "bcpandas"
__version__ = "1.3.0"


# BCP check
try:
    run(["bcp", "-v"], stdout=DEVNULL, stderr=DEVNULL, stdin=DEVNULL)
except FileNotFoundError:
    warnings.warn("BCP utility not installed or not found in PATH, bcpandas will not work!")

del run, DEVNULL, warnings



import urllib

import docker
from hypothesis import assume
from hypothesis.extra import pandas as hpd
import hypothesis.strategies as st
import pandas as pd
import pyodbc
import sqlalchemy as sa

from bcpandas.constants import _DELIMITER_OPTIONS, _QUOTECHAR_OPTIONS

# Hypo - typical use cases
#   - DataFrame: at least one row
#   - Text: All text in ASCII 32-127
#   - Integers: between -2**31-1 and 2**31-1
#   - Floats: between -2**31-1 and 2**31-1, without NaN or inf
#   - Dates
#   - Datetimes
#   - (some) booleans

MAX_VAL = 2 ** 31 - 1

# Strategies
strat_text = st.text(alphabet=st.characters(min_codepoint=32, max_codepoint=127), min_size=0)
strat_ints = st.integers(min_value=-MAX_VAL, max_value=MAX_VAL)
strat_floats = st.floats(
    min_value=-MAX_VAL, max_value=MAX_VAL, allow_nan=False, allow_infinity=False
)
strat_dates = st.dates()

strat_df_index = hpd.range_indexes(min_size=1)


df_hypo_mixed = hpd.data_frames(
    columns=[
        hpd.column(name="col1_text", elements=strat_text),
        hpd.column(name="col2_ints", elements=strat_ints),
        hpd.column(name="col3_floats", elements=strat_floats),
        hpd.column(name="col4_dates", elements=strat_dates),
        hpd.column(name="col4_bools", elements=st.booleans()),
    ],
    index=strat_df_index,
)

df_hypo_text = hpd.data_frames(columns=hpd.columns(5, elements=strat_text), index=strat_df_index)
df_hypo_ints = hpd.data_frames(columns=hpd.columns(5, elements=strat_ints), index=strat_df_index)
df_hypo_floats = hpd.data_frames(
    columns=hpd.columns(5, elements=strat_floats), index=strat_df_index
)
df_hypo_dates = hpd.data_frames(columns=hpd.columns(5, elements=strat_dates), index=strat_df_index)


def not_has_all_delims(df: pd.DataFrame) -> bool:
    return not all(
        df.applymap(lambda x: delim in x if isinstance(x, str) else False).any().any()
        for delim in _DELIMITER_OPTIONS
    )


def not_has_all_quotechars(df: pd.DataFrame) -> bool:
    return not all(
        df.applymap(lambda x: qc in x if isinstance(x, str) else False).any().any()
        for qc in _QUOTECHAR_OPTIONS
    )


def assume_not_all_delims_and_quotechars(df: pd.DataFrame) -> bool:
    return assume(not_has_all_delims(df) and not_has_all_quotechars(df))


def prep_df_for_comparison(df: pd.DataFrame, index: bool) -> pd.DataFrame:
    """
    Prepares a test dataframe for comparison with its corresponding data that is read from SQL.
    Becuase SQL does some implicit conversions, we need to make the df match that.

    Also uses the index=True/False param to set the df here to the expected value.
    """
    # if index=True, then make the df match the actual in SQL
    if index:
        df = df.reset_index()

    # SQL stores column names that are numbers as text, so we convert the dataframe columns to str
    # so that pandas-sql comparison will not fail
    try:
        df.columns = df.columns.astype(str)
    except TypeError:
        pass

    # Empty string becomes NULL in SQL (None in pandas), marking as ok for now
    if "object" in df.dtypes.values:
        df = df.replace({"": None})
    return df


def execute_sql_statement(sql_alchemy_engine: sa.engine.Engine, statement: str):
    """
    Executes the SQL statement using the provided engine. Assumes uses pyodbc.

    need to use pyodbc instead of sqlalchemy, otherwise get error:
    [SQL Server]CREATE DATABASE statement not allowed within multi-statement transaction.
    """
    conn = pyodbc.connect(sql_alchemy_engine.url.query["odbc_connect"], autocommit=True)
    conn.execute(statement)
    conn.close()


class DockerDB:
    """
    Class to create and run a SQL Server database using a Docker container. All the docker stuff is taken care of 
    by Python under the hood.

    Each instance of this class can only support a single specific combination of settings, such as container name
    and SQL password. To create more than one, create more instances of the class.
    """

    def __init__(
        self,
        container_name: str,
        sa_sql_password: str,
        mssql_image: str = "mcr.microsoft.com/mssql/server:2017-latest",
        port_host: int = 1433,
        port_container: int = 1433,
        accept_eula: bool = True,
        mssql_pid: str = "Express",
    ):
        self.client = docker.from_env()
        self.container_name = container_name
        self.sa_sql_password = sa_sql_password
        self.mssql_image = mssql_image
        self.port_host = port_host
        self.port_container = port_container
        self.accept_eula = accept_eula
        self.mssql_pid = mssql_pid

    def start(self):
        if not self.accept_eula:
            raise ValueError("Must accept Microsft's End User License Agreement")
        env = {
            "ACCEPT_EULA": "Y",
            "SA_PASSWORD": self.sa_sql_password,
        }
        if self.mssql_image.startswith("mcr.microsoft.com/mssql/server"):
            # means it's linux
            env["MSSQL_PID"] = self.mssql_pid

        self.container = self.client.containers.run(
            image=self.mssql_image,
            name=self.container_name,
            detach=True,  # '-d' flag
            environment=env,
            ports={self.port_container: self.port_host},
        )

    def stop(self):
        self.container.stop()

    def remove(self):
        self.container.remove()

    def create_engine(self, db_name="master") -> sa.engine.Engine:
        """Creates SQLAlchemy pyodbc engine for connecting to specified database (default master) as SA user"""
        db_url = (
            "Driver={ODBC Driver 17 for SQL Server};"
            + f"Server={self.address};Database={db_name};UID=sa;PWD={self.sa_sql_password};"
        )
        return sa.engine.create_engine(
            f"mssql+pyodbc:///?odbc_connect={urllib.parse.quote_plus(db_url)}"
        )

    def create_database(self, name):
        """Creates a SQL database on the SQL server"""
        execute_sql_statement(self.create_engine("master"), f"CREATE DATABASE {name}")

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()
        return

    @property
    def address(self):
        return f"127.0.0.1,{self.port_container}"



from hypothesis import assume, given, settings
from hypothesis.extra import pandas as hpd
import pandas as pd
from pandas.testing import assert_frame_equal
import pyodbc
import pytest

from bcpandas.constants import BCPandasValueError, read_data_settings
from bcpandas.tests.utils import (
    not_has_all_delims,
    not_has_all_quotechars,
    strat_dates,
    strat_floats,
    strat_ints,
    strat_text,
)

from .read_sql import read_sql

hypo_df = hpd.data_frames(
    columns=[
        hpd.column(name="col1_text", elements=strat_text),
        hpd.column(name="col2_ints", elements=strat_ints),
        hpd.column(name="col3_floats", elements=strat_floats),
        hpd.column(name="col4_dates", elements=strat_dates),
    ],
    index=hpd.range_indexes(min_size=1),
)


class TestReadSqlBasic:
    """
    For all tests, the 'actual' is retrieved using the built-in pandas methods, to compare to the 
    'expected' which used bcpandas. 

    Because dtypes change when reading from text files, ignoring dtypes checks. TODO how to really fix this
    """

    table_name = "lotr_readsql"
    view_name = f"v_{table_name}"

    @given(df=hypo_df)
    @settings(deadline=None)
    def test_readsql_table(self, df, sql_creds, database, pyodbc_creds):
        assume(not_has_all_delims(df))
        assume(not_has_all_quotechars(df))
        # create table and insert rows
        df.to_sql(
            name=self.table_name, con=pyodbc_creds, if_exists="replace", index=False, schema="dbo"
        )
        # get expected
        expected = read_sql(self.table_name, creds=sql_creds, sql_type="table", schema="dbo")
        # check
        assert_frame_equal(df, expected, check_dtype=False)

    @given(df=hypo_df)
    @settings(deadline=None)
    def test_readsql_view(self, df, sql_creds, database, pyodbc_creds):
        assume(not_has_all_delims(df))
        assume(not_has_all_quotechars(df))
        # create table and insert rows
        df.to_sql(
            name=self.table_name, con=pyodbc_creds, if_exists="replace", index=False, schema="dbo"
        )
        # create corresponding view
        conn = pyodbc.connect(pyodbc_creds.engine.url.query["odbc_connect"], autocommit=True)
        conn.execute(f"DROP VIEW IF EXISTS dbo.{self.view_name}")
        conn.execute(f"CREATE VIEW dbo.{self.view_name} AS SELECT * FROM dbo.{self.table_name}")
        conn.close()

        # get expected
        expected = read_sql(self.view_name, creds=sql_creds, sql_type="view", schema="dbo")
        # check
        assert_frame_equal(df, expected, check_dtype=False)

    @given(df=hypo_df)
    @settings(deadline=None)
    def test_readsql_query(self, df, sql_creds, database, pyodbc_creds):
        assume(not_has_all_delims(df))
        assume(not_has_all_quotechars(df))
        # create table and insert rows
        df.to_sql(
            name=self.table_name, con=pyodbc_creds, if_exists="replace", index=False, schema="dbo"
        )
        # get expected
        expected = read_sql(
            f"SELECT * FROM {self.table_name}", creds=sql_creds, sql_type="query", schema="dbo"
        )
        # check
        assert_frame_equal(df, expected, check_dtype=False)

    def test_readsql_custom_delimiter(self, sql_creds, database, pyodbc_creds):
        df = pd.DataFrame(
            {
                "col1": ["Sam and <", "Frodo", "Merry"],
                "col4": [2107, 2108, 2109],  # integers
                "col5": [1.5, 2.5, 3],  # floats
            }
        )
        # create table and insert rows
        df.to_sql(
            name=self.table_name, con=pyodbc_creds, if_exists="replace", index=False, schema="dbo"
        )
        # get expected
        expected = read_sql(
            self.table_name, creds=sql_creds, sql_type="table", schema="dbo", delimiter="|"
        )
        # check
        assert_frame_equal(df, expected)

        # check that correctly finds the error if bad cust delim is passed
        with pytest.raises(BCPandasValueError):
            read_sql(
                self.table_name, creds=sql_creds, sql_type="table", schema="dbo", delimiter="<"
            )

    def test_readsql_bad_delimiter(self, sql_creds, database, pyodbc_creds):
        # get default delimiter
        delim_default = read_data_settings["delimiter"]
        # has comma in data field, which is also the default delimiter
        df = pd.DataFrame(
            {
                "col1": [f"Sam and {delim_default}", "Frodo", "Merry"],
                "col4": [2107, 2108, 2109],  # integers
            }
        )
        # create table and insert rows
        df.to_sql(
            name=self.table_name, con=pyodbc_creds, if_exists="replace", index=False, schema="dbo"
        )
        # check that correctly finds the error
        with pytest.raises(BCPandasValueError):
            read_sql(self.table_name, creds=sql_creds, sql_type="table", schema="dbo")

        # check other error occurs if don't check_delim
        # TODO which error does it raise?
        expected = read_sql(
            self.table_name, creds=sql_creds, sql_type="table", schema="dbo", check_delim=False
        )
        assert_frame_equal(df, expected)



from contextlib import contextmanager  # noqa: E999
import json
from math import floor
import platform
from subprocess import PIPE, run
import sys
import time
from typing import Dict, List, Union

import click
from codetiming import Timer
import numpy as np
import pandas as pd
from read_sql.read_sql import read_sql

from bcpandas import SqlCreds, to_sql
from bcpandas.tests.utils import DockerDB

mssql_image = "mcr.microsoft.com/mssql/server:2017-latest"
_IS_WIN32 = sys.platform == "win32"
with_shell = False
if not _IS_WIN32:
    with_shell = True


def _parse_cmd(cmd: List[str]) -> Union[List[str], str]:
    if _IS_WIN32:
        return cmd
    else:
        return " ".join(cmd)


@contextmanager
def capture_stdout():
    # per https://stackoverflow.com/a/1218951/6067848
    from io import StringIO

    try:
        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()
        yield mystdout
    finally:
        sys.stdout = old_stdout


def gather_env_info():
    # TODO get RAM and hard drive info
    config = {
        "os_hardware": {},
        "python": {},
        "python_libs": {},
        "mssql": {},
        "mssql_tools": {},
        "docker": {},
    }

    # OS and hardware info
    config["os_hardware"] = {
        i: getattr(platform.uname(), i)
        for i in ["system", "release", "version", "machine", "processor"]
    }
    config["os_hardware"]["python_compiler"] = platform.python_compiler()

    # Python
    config["python"] = {k: v for k, v in sys.implementation.__dict__.items()}

    # Python libraries

    with capture_stdout() as mystdout:
        pd.show_versions(as_json=True)
        pdv = mystdout.getvalue().replace("'", '"').replace("None", "null")
        pdv_j = json.loads(pdv)
    config["python_libs"]["pandas_versions"] = pdv_j

    # MSSQL
    config["mssql"] = {"docker-image": mssql_image, "MSSQL_PID": "Express"}

    # Sql Tools
    cmd_bcp = ["bcp", "-v"]
    res = run(_parse_cmd(cmd_bcp), stdout=PIPE, stderr=PIPE, shell=with_shell)
    if res.returncode == 0:
        config["mssql_tools"] = {"bcp-version": res.stdout.decode().strip().split("\r\n")}

    # Docker
    cmd_docker = ["docker", "version", "--format", "'{{json .}}'"]
    res = run(_parse_cmd(cmd_docker), stdout=PIPE, stderr=PIPE, shell=with_shell)
    if res.returncode == 0:
        docker_out = res.stdout.decode().strip()[1:-1]  # strip outer single quotes
        config["docker"] = {"docker-version-output": json.loads(docker_out)}

    return config


def setup(docker_db):
    # start docker container
    docker_db.start()
    time.sleep(20)  # wait for container to start up

    # create database
    docker_db.create_database("benchmark_db")

    # create creds
    creds = SqlCreds.from_engine(docker_db.create_engine(db_name="benchmark_db"))

    return creds


def teardown(docker_db):
    docker_db.stop()
    docker_db.remove()


def _run_single_func(title, func, **kwargs):
    print(f"starting {title}")
    t = Timer(name=title)
    t.start()
    func(**kwargs)
    elapsed = t.stop()
    print(f"finished {title}")
    return elapsed


def run_benchmark_tosql(df: pd.DataFrame, creds: SqlCreds) -> Dict[str, float]:

    # using multi-insert in MS SQL is limited by hard limit of 2100 params
    # in SQL SPs. Using 2000 to be safe.
    # https://stackoverflow.com/a/56583204/6067848
    chunk_size = floor(2000 / (len(df.columns) + 1))  # +1 in case index=True

    funcs = [
        dict(
            title=f"pandas_multiinsert_{chunk_size}",
            func=df.to_sql,
            name="tbl_pandas_1",
            con=creds.engine,
            if_exists="replace",
            method="multi",
            chunksize=chunk_size,
        ),
        dict(
            title=f"bcpandas_batchsize_{chunk_size}",
            func=to_sql,
            df=df,
            table_name="tbl_bcpandas_1",
            creds=creds,
            if_exists="replace",
            batch_size=chunk_size,
        ),
        dict(
            title="bcpandas_batchsize_10000",
            func=to_sql,
            df=df,
            table_name="tbl_bcpandas_2",
            creds=creds,
            if_exists="replace",
            batch_size=10000,
        ),
    ]

    return {i["title"]: _run_single_func(**i) for i in funcs}


def run_benchmark_readsql(df: pd.DataFrame, creds: SqlCreds) -> Dict[str, float]:
    chunk_size = floor(2000 / (len(df.columns) + 1))  # +1 in case index=True

    # first create table and insert rows
    tbl_name = "sql_tbl_read_sql"
    to_sql(
        df,
        table_name=tbl_name,
        creds=creds,
        sql_type="table",
        schema="dbo",
        index=False,
        if_exists="replace",
        batch_size=10_000,
    )

    funcs = [
        dict(
            title=f"pandas_readsql_{chunk_size}",
            func=pd.read_sql_table,
            schema="dbo",
            table_name=tbl_name,
            con=creds.engine,
            chunksize=chunk_size,
        ),
        dict(
            title=f"bcpandas_batchsize_{chunk_size}_check_delim",
            func=read_sql,
            table_name=tbl_name,
            creds=creds,
            sql_type="table",
            schema="dbo",
            batch_size=chunk_size,
            check_delim=True,
        ),
        dict(
            title=f"bcpandas_batchsize_{chunk_size}_no_check_delim",
            func=read_sql,
            table_name=tbl_name,
            creds=creds,
            sql_type="table",
            schema="dbo",
            batch_size=chunk_size,
            check_delim=False,
        ),
    ]

    return {i["title"]: _run_single_func(**i) for i in funcs}


def save_and_plot(func, results, num_cols):
    # file names
    data_file = f"{func}_benchmark_data.json"
    plot_file = f"{func}_benchmark.png"
    env_file = f"{func}_benchmark_environment.json"

    # process results
    frame = pd.DataFrame(results).set_index("num_rows")
    frame.to_json(data_file)
    read_or_write = func.split("sql")[0]
    plot = frame.plot(
        kind="line",
        title=f"{read_or_write}_sql Comparison - Integers, {num_cols} columns",
        linestyle="--",
        marker="o",
    )
    plot.set_xlabel("number of rows")
    plot.set_ylabel("time (in seconds)")
    # https://stackoverflow.com/a/44444489/6067848
    plot.set_xticklabels([f"{x:,.0f}" for x in plot.get_xticks()])
    plot.get_figure().savefig(plot_file)
    env_info = gather_env_info()
    with open(env_file, "wt") as file:
        json.dump(env_info, file, indent=2)


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "-f",
    "--func",
    type=click.Choice(["tosql", "readsql"], case_sensitive=False),
    required=True,
    help="The Bcpandas function to benchmark",
)
@click.option(
    "--num-cols", type=int, default=6, show_default=True, help="Number of columns in the DataFrames"
)
@click.option(
    "--min-rows",
    type=click.IntRange(0,),
    default=50_000,
    show_default=True,
    help="Min rows in the DataFrames",
)
@click.option(
    "--max-rows", type=int, default=1_000_000, show_default=True, help="Max rows in the DataFrames"
)
@click.option(
    "--num-examples", type=int, default=10, show_default=True, help="How many Dataframes to run"
)
def main(func, num_cols, min_rows, max_rows, num_examples):
    """
    Will generate `num-examples` of DataFrames using numpy.linspace, going from `min-rows` rows to
    `max-rows` rows.
    """
    bmark_name = f"Benchmark run: func={func}, num_cols={num_cols}, min_rows={min_rows}, max_rows={max_rows}, num_examples={num_examples}"

    print(f"Starting {bmark_name}")
    timer = Timer(name=bmark_name)
    timer.start()
    docker_db = DockerDB("bcpandas-benchmarks", "MyBigSQLPasswordAlso!!!")
    try:
        # run benchmarks
        creds = setup(docker_db)
        results = []
        for n in np.linspace(min_rows, max_rows, num=num_examples):
            num_rows = int(n)
            df = pd.DataFrame(
                data=np.ndarray(shape=(num_rows, num_cols), dtype=int),
                columns=[f"col-{x}" for x in range(num_cols)],
            )
            if func == "readsql":
                _results = run_benchmark_readsql(df=df, creds=creds)
            elif func == "tosql":
                _results = run_benchmark_tosql(df=df, creds=creds)
            results.append({"num_rows": num_rows, **_results})
    finally:
        teardown(docker_db)

    save_and_plot(func=func, results=results, num_cols=num_cols)


if __name__ == "__main__":
    cli()



import pytest

from bcpandas.utils import run_cmd


@pytest.mark.parametrize(
    "cmd,use_print,expected_stdout,expected_stderr",
    [
        (["echo", "hello world from 1"], True, "hello world from 1\n", ""),
        pytest.param(
            ["python", "-c", "\"import sys; print('world2', file=sys.stderr)\""],
            True,
            "",
            "world2\n",
            marks=pytest.mark.skip(reason="Can't get it to write to STDERR for some reason"),
        ),
        (["echo", "hello world from 3"], False, "", ""),
        pytest.param(
            ["python", "-c", "\"import sys; print('world2', file=sys.stderr)\""],
            False,
            "",
            "",
            marks=pytest.mark.skip(reason="Can't get it to write to STDERR for some reason"),
        ),
    ],
)
# https://docs.pytest.org/en/6.2.x/capture.html
def test_run_cmd_print_output(capsys, cmd, use_print, expected_stdout, expected_stderr):
    run_cmd(cmd, print_output=use_print)
    captured = capsys.readouterr()
    assert captured.out == expected_stdout
    assert captured.err == expected_stderr



# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 23:36:07 2019

@author: ydima


There are 2 categories of tests we want to do:
    - Test every code path, i.e. every combination of arguments for each function
    - Test with different datasets that have different properties
"""

from datetime import date
from os.path import expandvars
from pathlib import Path
import sys
from typing import Optional, no_type_check

from hypothesis import HealthCheck, given, settings
import hypothesis.strategies as st
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
import pytest
import sqlalchemy

from bcpandas import to_sql
from bcpandas.constants import _DELIMITER_OPTIONS, _QUOTECHAR_OPTIONS, BCPandasValueError

from .utils import (
    assume_not_all_delims_and_quotechars,
    df_hypo_dates,
    df_hypo_floats,
    df_hypo_ints,
    df_hypo_mixed,
    df_hypo_text,
    execute_sql_statement,
    prep_df_for_comparison,
)


@pytest.mark.usefixtures("database")
def test_tosql_all_delims(sql_creds):
    df = pd.DataFrame(
        {i: [v, "random", "string", 1, 2.0] for i, v in enumerate(_DELIMITER_OPTIONS)}
    )
    with pytest.raises(BCPandasValueError):
        to_sql(df=df, table_name="tbl_all_delims", creds=sql_creds, if_exists="replace")


@pytest.mark.usefixtures("database")
def test_tosql_all_quotechars(sql_creds):
    df = pd.DataFrame(
        {i: [v, "random", "string", 1, 2.0] for i, v in enumerate(_QUOTECHAR_OPTIONS)}
    )
    with pytest.raises(BCPandasValueError):
        to_sql(df=df, table_name="tbl_all_delims", creds=sql_creds, if_exists="replace")


def _get_bcp_path() -> Optional[str]:
    """
    On Windows, is typically in C:/Program Files/Microsoft SQL Server/Client SDK/ODBC/170/Tools/Binn
    On Linux it depends where you install it, but usually in /opt/mssql-tools/bin
    """
    try:
        if sys.platform == "win32":
            first_part = (
                Path(expandvars("%ProgramFiles%")) / "Microsoft SQL Server" / "Client SDK" / "ODBC"
            )
            version = max(x.parts[-1] for x in first_part.iterdir())
            bcp_path = first_part / version / "Tools" / "Binn"
            return str(bcp_path / "bcp.exe")
        else:
            bcp_path = Path("/opt/mssql-tools/bin")
            if bcp_path.resolve().exists():
                return str(bcp_path / "bcp")
            else:
                return None
    except FileNotFoundError:
        return None


_bcp_path = _get_bcp_path()


@pytest.mark.parametrize(
    "bcp_path",
    [
        pytest.param(
            _bcp_path,
            id=_bcp_path,
            marks=[
                pytest.mark.skipif(
                    _bcp_path is None, reason=f"Didn't find full path to BCP on {sys.platform}"
                )
            ],
        )
    ],
)
@pytest.mark.usefixtures("database")
def test_tosql_full_bcp_path(bcp_path, sql_creds):
    df = pd.DataFrame({"col1": ["a", "b", "c", "d"], "col2": [1.5, 2.5, 3.5, 4.5]})
    tbl_name = "tbl_df_full_bcp_path"
    schema_name = "dbo"
    execute_sql_statement(sql_creds.engine, f"DROP TABLE IF EXISTS {schema_name}.{tbl_name}")
    to_sql(
        df=df,
        table_name=tbl_name,
        creds=sql_creds,
        schema=schema_name,
        if_exists="replace",
        index=False,
        bcp_path=bcp_path,
    )

    # check result
    actual = pd.read_sql_query(sql=f"SELECT * FROM {schema_name}.{tbl_name}", con=sql_creds.engine)
    expected = prep_df_for_comparison(df=df, index=False)
    assert_frame_equal(expected, actual)


@pytest.mark.usefixtures("database")
@no_type_check  # gives wierd errors in list values of dfs in params
@pytest.mark.parametrize(
    "df",
    [
        pytest.param(
            pd.DataFrame(
                {
                    "col1": ["a", "b", "c", "d"],
                    "col2": [1, np.NaN, 3, np.NaN],
                    "col3": [1.5, 2.5, 3.5, 4.5],
                }
            ),
            id="nan",
        ),
        pytest.param(
            pd.DataFrame(
                {
                    "col1": ["a", "b", "c", "d"],
                    "col2": [1, 2, 3, 4],
                    "col3": [1.5, np.NaN, 3.5, np.NaN],
                }
            ),
            id="nan_last_col",
        ),
        pytest.param(
            pd.DataFrame(
                {"col1": [1.5, 2.5, 3.5, 4.5], "col2": [1, 2, 3, 4], "col3": ["a", None, "c", None]}
            ),
            id="None_last_col",
        ),
        pytest.param(
            pd.DataFrame(
                {
                    "col1": [1.5, np.inf, 3.5, 4.5],
                    "col2": [1, 2, np.inf, 4],
                    "col3": ["a", "b", "c", None],
                }
            ),
            id="inf",
            marks=pytest.mark.xfail,
        ),
    ],
)
def test_tosql_nan_null_inf(df, sql_creds):
    tbl_name = "tbl_df_nan_null_last_col"
    schema_name = "dbo"
    execute_sql_statement(sql_creds.engine, f"DROP TABLE IF EXISTS {schema_name}.{tbl_name}")
    to_sql(
        df=df,
        table_name=tbl_name,
        creds=sql_creds,
        schema=schema_name,
        if_exists="replace",
        index=False,
    )

    # check result
    actual = pd.read_sql_query(sql=f"SELECT * FROM {schema_name}.{tbl_name}", con=sql_creds.engine)
    expected = prep_df_for_comparison(df=df, index=False)
    assert_frame_equal(expected, actual)


@pytest.mark.usefixtures("database")
@pytest.mark.parametrize(
    "df",
    [pd.DataFrame({}), pd.DataFrame({"a": []}), pd.DataFrame(index=["a"])],
    ids=["df_empty", "df_with_col", "df_with_idx"],
)
def test_tosql_empty_df(df, sql_creds):
    tbl_name = "tbl_df_empty"
    schema_name = "dbo"
    execute_sql_statement(sql_creds.engine, f"DROP TABLE IF EXISTS {schema_name}.{tbl_name}")
    to_sql(df=df, table_name=tbl_name, creds=sql_creds, schema=schema_name, if_exists="replace")
    # make sure nothing happened in the database
    qry = """
        SELECT * 
        FROM INFORMATION_SCHEMA.TABLES 
        WHERE TABLE_SCHEMA = '{_schema}' 
        AND TABLE_NAME = '{_tbl}'
        """.format(
        _tbl=tbl_name, _schema=schema_name
    )
    res = pd.read_sql_query(sql=qry, con=sql_creds.engine)
    # assert that rows == 0, it has columns even without rows because it is an internal system table
    assert res.shape[0] == 0


@pytest.mark.usefixtures("database")
def test_duplicate_columns(sql_creds):
    table_name = "tosql_column_scenarios_1"
    df = pd.DataFrame(
        {
            "col1": [1.5, 2.5, 3.5, 4.5],
            "col2": [5.5, 6.5, 7.5, 8.5],
            "col3": [9.5, 10.5, 11.5, 12.5],
            "col4": [13.5, 14.5, 15.5, 16.5],
        }
    )

    # set duplicate column
    df = df.rename(columns={"col3": "col2"})

    with pytest.raises(BCPandasValueError):
        to_sql(
            df=df,
            table_name=table_name,
            creds=sql_creds,
            if_exists="replace",
            index=False,
            sql_type="table",
        )


@pytest.mark.skip(reason="Didn't test yet")
def test_non_string_columns():
    # TODO
    pass


@pytest.mark.parametrize("batch_size", [0, 50])
def test_batch_size_param(sql_creds, batch_size):
    """
    Test that batch size can't be:
        - 0
        - larger than the number of rows
    """
    df = pd.DataFrame(
        {
            "col1": [1.5, 2.5, 3.5, 4.5],
            "col2": [5.5, 6.5, 7.5, 8.5],
            "col3": [9.5, 10.5, 11.5, 12.5],
            "col4": [13.5, 14.5, 15.5, 16.5],
        }
    )
    with pytest.raises(BCPandasValueError):
        to_sql(
            df=df,
            table_name="some_table",
            creds=sql_creds,
            if_exists="replace",
            index=False,
            sql_type="table",
            batch_size=batch_size,
        )


@pytest.mark.usefixtures("database")
class _BaseToSql:
    sql_type = "table"
    schema_name = "dbo"

    # per https://github.com/pytest-dev/pytest/issues/2618#issuecomment-318584202
    fixture_names = ("sql_creds", "pyodbc_creds")

    @pytest.fixture(autouse=True, scope="function")
    def auto_injector_fixture(self, request):
        for name in self.fixture_names:
            setattr(self, name, request.getfixturevalue(name))


class TestToSqlColumnScenarios(_BaseToSql):
    """
    Various tests for scenarios where the dataframe columns don't map exactly to the database columns,
    whether by different order, missing columns, etc.
    """

    table_name = "tosql_column_scenarios"
    df = pd.DataFrame(
        {
            "col1": [1.5, 2.5, 3.5, 4.5],
            "col2": [5.5, 6.5, 7.5, 8.5],
            "col3": [9.5, 10.5, 11.5, 12.5],
            "col4": [13.5, 14.5, 15.5, 16.5],
        }
    )

    def _create_tbl_with_data(self):
        # first create a new table and put some data into it
        execute_sql_statement(
            self.sql_creds.engine, f"DROP TABLE IF EXISTS {self.schema_name}.{self.table_name}"
        )
        to_sql(
            df=self.df,
            table_name=self.table_name,
            creds=self.sql_creds,
            if_exists="replace",
            index=False,
            sql_type=self.sql_type,
        )

    def _check_result(self, df_modified):
        # check result
        actual = pd.read_sql_query(
            sql=f"SELECT * FROM {self.schema_name}.{self.table_name}", con=self.sql_creds.engine
        )
        expected = prep_df_for_comparison(
            df=pd.concat([self.df, df_modified], axis=0, ignore_index=True, sort=False), index=False
        )
        assert_frame_equal(expected, actual)

    def test_skip_cols(self):
        """
        see https://docs.microsoft.com/en-us/sql/relational-databases/import-export/use-a-format-file-to-skip-a-table-column-sql-server
        """
        # first create a new table and put some data into it
        self._create_tbl_with_data()

        # then insert (append) data into only some of the columns
        df_changed = self.df.iloc[:, 1:]
        to_sql(
            df=df_changed,
            table_name=self.table_name,
            creds=self.sql_creds,
            if_exists="append",
            index=False,
            sql_type=self.sql_type,
        )

        # check result
        self._check_result(df_changed)

    def test_cols_out_of_order(self):
        """
        see https://docs.microsoft.com/en-us/sql/relational-databases/import-export/use-a-format-file-to-map-table-columns-to-data-file-fields-sql-server
        """
        # first create a new table and put some data into it
        self._create_tbl_with_data()

        # then insert (append) data with columns out of order
        df_changed = self.df.iloc[:, ::-1]  # columns reversed
        to_sql(
            df=df_changed,
            table_name=self.table_name,
            creds=self.sql_creds,
            if_exists="append",
            index=False,
            sql_type=self.sql_type,
        )

        # check result
        self._check_result(df_changed)

    def test_cols_skip_and_out_of_order(self):
        # first create a new table and put some data into it
        self._create_tbl_with_data()

        # then insert data into only some of the columns, and out of order
        df_changed = self.df.loc[:, ["col4", "col1", "col2"]]
        to_sql(
            df=df_changed,
            table_name=self.table_name,
            creds=self.sql_creds,
            if_exists="append",
            index=False,
            sql_type=self.sql_type,
        )

        # check result
        self._check_result(df_changed)

    def test_extra_col(self):
        # first create a new table and put some data into it
        self._create_tbl_with_data()

        # then insert data into only some of the columns, and out of order
        df_changed = self.df.copy()
        df_changed.loc[:, "col_N"] = [17.5, 18.5, 19.5, 20.5]

        with pytest.raises(BCPandasValueError):
            to_sql(
                df=df_changed,
                table_name=self.table_name,
                creds=self.sql_creds,
                if_exists="append",
                index=False,
                sql_type=self.sql_type,
            )


class TestToSqlReplace(_BaseToSql):
    """

    """

    table_name = "lotr_tosql_replace"

    def _test_base(self, df, index):
        assume_not_all_delims_and_quotechars(df)
        to_sql(
            df=df,
            table_name=self.table_name,
            creds=self.sql_creds,
            index=index,
            sql_type=self.sql_type,
            if_exists="replace",
        )
        actual = pd.read_sql_query(sql=f"SELECT * FROM {self.table_name}", con=self.pyodbc_creds)
        expected = prep_df_for_comparison(df=df, index=index)
        assert_frame_equal(expected, actual, check_column_type="equiv")

    def _test_exists(self, df, index):
        self._test_base(df, index)

    def _test_not_exists(self, df, index):
        execute_sql_statement(
            self.pyodbc_creds.engine, f"DROP TABLE IF EXISTS dbo.{self.table_name}"
        )
        self._test_base(df, index)

    @given(df=df_hypo_mixed, index=st.booleans())
    @settings(deadline=None)
    def test_df_mixed(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_text, index=st.booleans())
    @settings(deadline=None)
    def test_df_text(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_ints, index=st.booleans())
    @settings(deadline=None)
    def test_df_ints(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_floats, index=st.booleans())
    @settings(deadline=None)
    def test_df_floats(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_dates, index=st.booleans())
    @settings(deadline=None)
    def test_df_dates(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)


class TestToSqlAppend(_BaseToSql):
    """

    """

    table_name = "lotr_tosql_append"

    def _test_exists(self, df, index):
        assume_not_all_delims_and_quotechars(df)
        # first populate the data
        to_sql(
            df=df,
            table_name=self.table_name,
            creds=self.sql_creds,
            index=index,
            sql_type=self.sql_type,
            if_exists="replace",
        )

        # then test to sql with option of 'append'
        to_sql(
            df=df,
            table_name=self.table_name,
            creds=self.sql_creds,
            index=index,
            sql_type=self.sql_type,
            if_exists="append",
        )
        actual = pd.read_sql_query(
            sql=f"SELECT * FROM dbo.{self.table_name}", con=self.pyodbc_creds
        )
        df = prep_df_for_comparison(df=df, index=index)
        expected = pd.concat([df, df], axis=0, ignore_index=True)  # appended
        assert_frame_equal(expected, actual, check_column_type="equiv")

    def _test_not_exists(self, df, index):
        execute_sql_statement(
            self.pyodbc_creds.engine, f"DROP TABLE IF EXISTS dbo.{self.table_name}"
        )
        assume_not_all_delims_and_quotechars(df)
        # test to sql with option of 'append'
        to_sql(
            df=df,
            table_name=self.table_name,
            creds=self.sql_creds,
            index=index,
            sql_type=self.sql_type,
            if_exists="append",
        )
        actual = pd.read_sql_query(
            sql=f"SELECT * FROM dbo.{self.table_name}", con=self.pyodbc_creds
        )
        expected = prep_df_for_comparison(df=df, index=index)
        assert_frame_equal(expected, actual, check_column_type="equiv")

    @given(df=df_hypo_mixed, index=st.booleans())
    @settings(deadline=None)
    def test_df_mixed(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_text, index=st.booleans())
    @settings(deadline=None)
    def test_df_text(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_ints, index=st.booleans())
    @settings(deadline=None)
    def test_df_ints(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_floats, index=st.booleans())
    @settings(deadline=None)
    def test_df_floats(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_dates, index=st.booleans())
    @settings(deadline=None)
    def test_df_dates(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)


class TestToSqlFail(_BaseToSql):
    """

    """

    table_name = "lotr_tosql_fail"

    def _test_exists(self, df, index):
        assume_not_all_delims_and_quotechars(df)
        # first populate the data
        to_sql(
            df=df,
            table_name=self.table_name,
            creds=self.sql_creds,
            index=index,
            sql_type=self.sql_type,
            if_exists="replace",
        )

        # then test to sql with option of 'fail'
        with pytest.raises(BCPandasValueError):
            to_sql(
                df=df,
                table_name=self.table_name,
                creds=self.sql_creds,
                index=index,
                sql_type=self.sql_type,
                if_exists="fail",
            )

    def _test_not_exists(self, df, index):
        execute_sql_statement(
            self.pyodbc_creds.engine, f"DROP TABLE IF EXISTS dbo.{self.table_name}"
        )
        assume_not_all_delims_and_quotechars(df)
        to_sql(
            df=df,
            table_name=self.table_name,
            creds=self.sql_creds,
            index=index,
            sql_type=self.sql_type,
            if_exists="fail",
        )
        actual = pd.read_sql_query(sql=f"SELECT * FROM {self.table_name}", con=self.pyodbc_creds)
        expected = prep_df_for_comparison(df=df, index=index)
        assert_frame_equal(expected, actual, check_column_type="equiv")

    @given(df=df_hypo_mixed, index=st.booleans())
    @settings(deadline=None, suppress_health_check=[HealthCheck.too_slow])
    def test_df_mixed(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_text, index=st.booleans())
    @settings(deadline=None, suppress_health_check=[HealthCheck.too_slow])
    def test_df_text(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_ints, index=st.booleans())
    @settings(deadline=None, suppress_health_check=[HealthCheck.too_slow])
    def test_df_ints(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_floats, index=st.booleans())
    @settings(deadline=None, suppress_health_check=[HealthCheck.too_slow])
    def test_df_floats(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)

    @given(df=df_hypo_dates, index=st.booleans())
    @settings(deadline=None, suppress_health_check=[HealthCheck.too_slow])
    def test_df_dates(self, df, index):
        self._test_exists(df, index)
        self._test_not_exists(df, index)


class TestToSqlDtypeScenarios(_BaseToSql):
    """
    Various tests for scenarios where dtype is and is not passed, to check that the columns are the correct types
    """

    table_name = "tosql_dtype_scenarios"
    df = pd.DataFrame(
        {
            "col1": [1, 2, 3, 4],
            "col2": [5.5, 6.5, 7.5, 8.5],
            "col3": ["foo", "bar", "baz", "qux"],
            "col4": [date(2021, 1, 1), date(2021, 1, 2), date(2021, 1, 3), date(2021, 1, 4)],
            "col5": [True, True, False, False],
        }
    )

    def test_no_dtype(self):
        to_sql(
            df=self.df,
            table_name=self.table_name,
            creds=self.sql_creds,
            index=False,
            sql_type=self.sql_type,
            if_exists="replace",
            dtype=None,
        )

        actual = pd.read_sql_query(
            sql=f"""
            SELECT
              COLUMN_NAME,
              DATA_TYPE,
              CHARACTER_MAXIMUM_LENGTH
            FROM
              INFORMATION_SCHEMA.COLUMNS 
            WHERE
              TABLE_NAME = '{self.table_name}'""",
            con=self.pyodbc_creds,
        )
        expected = pd.DataFrame(
            {
                "COLUMN_NAME": ["col1", "col2", "col3", "col4", "col5"],
                "DATA_TYPE": ["bigint", "float", "varchar", "date", "bit"],
                "CHARACTER_MAXIMUM_LENGTH": [np.NaN, np.NaN, -1.0, np.NaN, np.NaN],
            }
        )
        assert_frame_equal(expected, actual)

    def test_with_dtype(self):
        to_sql(
            df=self.df,
            table_name=self.table_name,
            creds=self.sql_creds,
            index=False,
            sql_type=self.sql_type,
            if_exists="replace",
            dtype={
                "col1": sqlalchemy.types.INTEGER(),
                "col2": sqlalchemy.types.FLOAT(),
                "col3": sqlalchemy.types.NVARCHAR(length=10),
                "col4": sqlalchemy.types.DATE(),
                "col5": sqlalchemy.dialects.mssql.BIT(),
            },
        )

        actual = pd.read_sql_query(
            sql=f"""
            SELECT
              COLUMN_NAME,
              DATA_TYPE,
              CHARACTER_MAXIMUM_LENGTH
            FROM
              INFORMATION_SCHEMA.COLUMNS 
            WHERE
              TABLE_NAME = '{self.table_name}'""",
            con=self.pyodbc_creds,
        )
        expected = pd.DataFrame(
            {
                "COLUMN_NAME": ["col1", "col2", "col3", "col4", "col5"],
                "DATA_TYPE": ["int", "float", "nvarchar", "date", "bit"],
                "CHARACTER_MAXIMUM_LENGTH": [np.NaN, np.NaN, 10.0, np.NaN, np.NaN],
            }
        )
        assert_frame_equal(expected, actual)


class TestToSqlOther(_BaseToSql):
    """

    """

    table_name = "lotr_tosql_other"

    def _test_df_template(self, df, sql_creds, index):
        assume_not_all_delims_and_quotechars(df)
        with pytest.raises(AssertionError):
            to_sql(
                df=df,
                table_name=self.table_name,
                creds=self.sql_creds,
                index=index,
                sql_type=self.sql_type,
                if_exists="bad_arg",
            )

    @given(df=df_hypo_mixed, index=st.booleans())
    @settings(deadline=None)
    def test_df_mixed(self, df, sql_creds, index):
        self._test_df_template(df, sql_creds, index)

    @given(df=df_hypo_text, index=st.booleans())
    @settings(deadline=None)
    def test_df_text(self, df, sql_creds, index):
        self._test_df_template(df, sql_creds, index)

    @given(df=df_hypo_ints, index=st.booleans())
    @settings(deadline=None)
    def test_df_ints(self, df, sql_creds, index):
        self._test_df_template(df, sql_creds, index)

    @given(df=df_hypo_floats, index=st.booleans())
    @settings(deadline=None)
    def test_df_floats(self, df, sql_creds, index):
        self._test_df_template(df, sql_creds, index)

    @given(df=df_hypo_dates, index=st.booleans())
    @settings(deadline=None)
    def test_df_dates(self, df, sql_creds, index):
        self._test_df_template(df, sql_creds, index)



# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 23:07:15 2019

@author: ydima
"""

import csv
import logging
import os
from textwrap import dedent
from typing import Dict, Optional, Union
from urllib.parse import quote_plus

import pandas as pd
from pandas.io.sql import SQLDatabase, SQLTable
import sqlalchemy as sa

from .constants import (
    IF_EXISTS_OPTIONS,
    IN,
    NEWLINE,
    TABLE,
    BCPandasValueError,
    get_delimiter,
    get_quotechar,
)
from .utils import bcp, build_format_file, get_temp_file

logger = logging.getLogger(__name__)


class SqlCreds:
    """
    Credential object for all SQL operations. Will automatically also create a SQLAlchemy 
    engine that uses `pyodbc` as the DBAPI, and store it in the `self.engine` attribute.

    If `username` and `password` are not provided, `with_krb_auth` will be `True`.

    Only supports SQL based logins, not Active Directory or Azure AD.

    Parameters
    ----------
    server : str
    database : str
    username : str, optional
    password : str, optional
    driver_version : int, default 17
        The version of the Microsoft ODBC Driver for SQL Server to use 
    odbc_kwargs : dict of {str, str or int}, optional
        additional keyword arguments, to pass into ODBC connection string, 
        such as Encrypted='yes'
    
    Returns
    -------
    `bcpandas.SqlCreds`
    """

    def __init__(
        self,
        server: str,
        database: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        driver_version: int = 17,
        port: int = 1433,
        odbc_kwargs: Optional[Dict[str, Union[str, int]]] = None,
    ):
        self.server = server
        self.database = database
        self.port = port

        self.driver = f"{{ODBC Driver {driver_version} for SQL Server}}"

        # Append a comma for use in connection strings (optionally blank)
        if port:
            port_str = f",{self.port}"
        else:
            port_str = ""

        db_url = (
            f"Driver={self.driver};Server=tcp:{self.server}{port_str};Database={self.database};"
        )
        if username and password:
            self.username = username
            self.password = password
            self.with_krb_auth = False
            db_url += f"UID={username};PWD={password}"
        else:
            self.username = ""
            self.password = ""
            self.with_krb_auth = True
            db_url += "Trusted_Connection=yes;"

        logger.info(f"Created creds:\t{self}")

        # construct the engine for sqlalchemy
        if odbc_kwargs:
            db_url += ";".join(f"{k}={v}" for k, v in odbc_kwargs.items())
        conn_string = f"mssql+pyodbc:///?odbc_connect={quote_plus(db_url)}"
        self.engine = sa.engine.create_engine(conn_string)

        logger.info(f"Created engine for sqlalchemy:\t{self.engine}")

    @classmethod
    def from_engine(cls, engine: sa.engine.base.Engine) -> "SqlCreds":
        """
        Alternate constructor, from a `sqlalchemy.engine.base.Engine` that uses `pyodbc` as the DBAPI 
        (which is the SQLAlchemy default for MS SQL) and using an exact PyODBC connection string (not DSN or hostname).
        See https://docs.sqlalchemy.org/en/13/dialects/mssql.html#connecting-to-pyodbc for more.
        
        Parameters
        ----------
        engine : `sqlalchemy.engine.base.Engine`
            The SQLAlchemy engine object, configured as described above

        Returns
        -------
        `bcpandas.SqlCreds`
        """
        try:
            # get the odbc url part from the engine, split by ';' delimiter
            conn_url = engine.url.query["odbc_connect"].split(";")
            # convert into dict
            conn_dict = {x.split("=")[0]: x.split("=")[1] for x in conn_url if "=" in x}

            if "," in conn_dict["Server"]:
                conn_dict["port"] = int(conn_dict["Server"].split(",")[1])

            sql_creds = cls(
                server=conn_dict["Server"].replace("tcp:", "").split(",")[0],
                database=conn_dict["Database"],
                username=conn_dict.get("UID", None),
                password=conn_dict.get("PWD", None),
                port=conn_dict.get("port", None),
            )

            # add Engine object as attribute
            sql_creds.engine = engine
            return sql_creds
        except (KeyError, AttributeError) as ex:
            raise BCPandasValueError(
                f"The supplied 'engine' object could not be parsed correctly, try creating a SqlCreds object manually."
                f"\nOriginal Error: \n {ex}"
            )

    def __repr__(self):
        # adopted from https://github.com/erdewit/ib_insync/blob/master/ib_insync/objects.py#L51
        clsName = self.__class__.__qualname__
        kwargs = ", ".join(f"{k}={v!r}" for k, v in self.__dict__.items() if k != "password")
        if hasattr(self, "password"):
            kwargs += ", password=[REDACTED]"
        return f"{clsName}({kwargs})"

    __str__ = __repr__


def _sql_item_exists(sql_type: str, schema: str, table_name: str, creds: SqlCreds) -> bool:
    _qry = dedent(
        """
        SELECT * 
        FROM INFORMATION_SCHEMA.{_typ}S 
        WHERE TABLE_SCHEMA = '{_schema}' 
        AND TABLE_NAME = '{_tbl}'
        """.format(
            _typ=sql_type.upper(), _schema=schema, _tbl=table_name
        )
    )
    res = pd.read_sql_query(sql=_qry, con=creds.engine)
    return res.shape[0] > 0


def _create_table(
    schema: str,
    table_name: str,
    creds: SqlCreds,
    df: pd.DataFrame,
    if_exists: str,
    dtype: dict = None,
):
    """use pandas' own code to create the table and schema"""

    sql_db = SQLDatabase(engine=creds.engine, schema=schema)
    table = SQLTable(
        table_name,
        sql_db,
        frame=df,
        index=False,  # already set as new col earlier if index=True
        if_exists=if_exists,
        index_label=None,
        schema=schema,
        dtype=dtype,
    )
    table.create()


def _handle_cols_for_append(
    df: pd.DataFrame,
    table_name: str,
    creds: SqlCreds,
    sql_item_exists: bool,
    schema: str,
    if_exists: str,
):
    cols_dict = None
    if if_exists == "append":
        # get dict of column names -> order of column
        cols_dict = dict(
            pd.read_sql_query(
                dedent(
                    """
                SELECT COLUMN_NAME, ORDINAL_POSITION 
                FROM INFORMATION_SCHEMA.COLUMNS 
                WHERE TABLE_SCHEMA = '{_schema}'
                AND TABLE_NAME = '{_tbl}'
            """.format(
                        _schema=schema, _tbl=table_name
                    )
                ),
                creds.engine,
            ).values
        )

        # check that column names match in db and dataframe exactly
        if sql_item_exists:
            # the db cols are always strings, unlike df cols
            extra_cols = [str(x) for x in df.columns if str(x) not in cols_dict.keys()]
            if extra_cols:
                raise BCPandasValueError(
                    f"Column(s) detected in the dataframe that are not in the database, "
                    f"cannot have new columns if `if_exists=='append'`, "
                    f"the extra column(s): {extra_cols}"
                )
    return cols_dict


def _prepare_table(
    df: pd.DataFrame,
    table_name: str,
    creds: SqlCreds,
    sql_item_exists: bool,
    sql_type: str,
    schema: str,
    if_exists: str,
    dtype: Optional[dict],
) -> None:
    """
    Prepares the destination SQL table, handling the `if_exists` param.
    """
    if if_exists == "fail":
        if sql_item_exists:
            raise BCPandasValueError(
                f"The {sql_type} called {schema}.{table_name} already exists, "
                f"`if_exists` param was set to `fail`."
            )
        else:
            _create_table(
                schema=schema,
                table_name=table_name,
                creds=creds,
                df=df,
                if_exists=if_exists,
                dtype=dtype,
            )
    elif if_exists == "replace":
        _create_table(
            schema=schema,
            table_name=table_name,
            creds=creds,
            df=df,
            if_exists=if_exists,
            dtype=dtype,
        )
    elif if_exists == "append":
        if not sql_item_exists:
            _create_table(
                schema=schema,
                table_name=table_name,
                creds=creds,
                df=df,
                if_exists=if_exists,
                dtype=dtype,
            )


def _validate_args(
    df: pd.DataFrame, sql_type: str, if_exists: str, batch_size: Optional[int],
) -> None:
    assert sql_type == TABLE, "only supporting table, not view, for now"
    assert if_exists in IF_EXISTS_OPTIONS

    if df.columns.has_duplicates:
        raise BCPandasValueError(
            "Columns with duplicate names detected, SQL requires that column names be unique. "
            f"Duplicates: {df.columns[df.columns.duplicated(keep=False)]}"
        )

    if batch_size is not None:
        if batch_size == 0:
            raise BCPandasValueError("Param batch_size can't be 0")
        if batch_size > df.shape[0]:
            raise BCPandasValueError(
                "Param batch_size can't be larger than the number of rows in the DataFrame"
            )


def to_sql(
    df: pd.DataFrame,
    table_name: str,
    creds: SqlCreds,
    sql_type: str = "table",
    schema: str = "dbo",
    index: bool = True,
    if_exists: str = "fail",
    batch_size: int = None,
    debug: bool = False,
    bcp_path: str = None,
    dtype: dict = None,
    process_dest_table: bool = True,
    print_output: bool = True,
):
    """
    Writes the pandas DataFrame to a SQL table or view.

    Will write all columns to the table or view. If the destination table/view doesn't exist, will create it.
    Assumes the SQL table/view has the same number, name, and type of columns.
    To only write parts of the DataFrame, filter it beforehand and pass that to this function.
    Unlike the pandas counterpart, if the DataFrame has no rows, nothing will happen.

    Parameters
    ----------
    df : pandas.DataFrame
    table_name : str
        Name of SQL table or view, without the schema
    creds : bcpandas.SqlCreds
        The credentials used in the SQL database.
    sql_type : {'table'}, can only be 'table'
        The type of SQL object of the destination.
    schema : str, default 'dbo'
        The SQL schema.
    index : bool, default True
        Write DataFrame index as a column. Uses the index name as the column
        name in the table.
    if_exists : {'fail', 'replace', 'append'}, default 'fail'
        How to behave if the table already exists.
        * fail: Raise a BCPandasValueError.
        * replace: Drop the table before inserting new values.
        * append: Insert new values to the existing table. Matches the dataframe columns to the database columns by name.
            If the database table exists then the dataframe cannot have new columns that aren't in the table, 
            but conversely table columns can be missing from the dataframe.

    batch_size : int, optional
        Rows will be written in batches of this size at a time. By default, BCP sets this to 1000.
    debug : bool, default False
        If True, will not delete the temporary CSV and format files, and will output their location.
    bcp_path : str, default None
        The full path to the BCP utility, useful if it is not in the PATH environment variable
    dtype: dict, default None
        A dict with keys the names of columns and values SqlAlchemy types for defining their types. These are
        directly passed into pandas' API
    process_dest_table: bool, default True
        Internal: Only to be used when calling directly from within pandas using the `sql_engine` param.
        If False, then will skip preparing the destination table via the `if_exists` param,
        and will only attempt to insert data.
        You should generally not set this param yourself.
    print_output: bool, default True
        Whether to print output to STDOUT in real time. Regardless, the output will be logged.
        Added in version 1.3
    """
    # validation
    if df.shape[0] == 0 or df.shape[1] == 0:
        return

    _validate_args(df=df, sql_type=sql_type, if_exists=if_exists, batch_size=batch_size)

    # TODO diff way to implement? could be big performance hit with big dataframe
    if index:
        df = df.copy(deep=True).reset_index()

    delim = get_delimiter(df)
    quotechar = get_quotechar(df)

    # save to temp path
    csv_file_path = get_temp_file()
    # replace bools with 1 or 0, this is what pandas native does when writing to SQL Server
    df.replace({True: 1, False: 0}).to_csv(
        path_or_buf=csv_file_path,
        sep=delim,
        header=False,
        index=False,  # already set as new col earlier if index=True
        quoting=csv.QUOTE_MINIMAL,  # pandas default
        quotechar=quotechar,
        line_terminator=NEWLINE,
        doublequote=True,
        escapechar=None,  # not needed, as using doublequote
    )
    logger.debug(f"Saved dataframe to temp CSV file at {csv_file_path}")

    # build format file
    fmt_file_path = get_temp_file()

    sql_item_exists = _sql_item_exists(
        sql_type=sql_type, schema=schema, table_name=table_name, creds=creds
    )

    cols_dict = _handle_cols_for_append(
        df=df,
        table_name=table_name,
        creds=creds,
        sql_item_exists=sql_item_exists,
        schema=schema,
        if_exists=if_exists,
    )

    fmt_file_txt = build_format_file(df=df, delimiter=delim, db_cols_order=cols_dict)
    with open(fmt_file_path, "w") as ff:
        ff.write(fmt_file_txt)
    logger.debug(f"Created BCP format file at {fmt_file_path}")

    try:
        if process_dest_table:
            _prepare_table(
                df=df,
                table_name=table_name,
                creds=creds,
                sql_item_exists=sql_item_exists,
                sql_type=sql_type,
                schema=schema,
                if_exists=if_exists,
                dtype=dtype,
            )

        # BCP the data in
        bcp(
            sql_item=table_name,
            direction=IN,
            flat_file=csv_file_path,
            format_file_path=fmt_file_path,
            creds=creds,
            print_output=print_output,
            sql_type=sql_type,
            schema=schema,
            batch_size=batch_size,
            bcp_path=bcp_path,
        )
    finally:
        if not debug:
            logger.debug("Deleting temp CSV and format files")
            os.remove(csv_file_path)
            os.remove(fmt_file_path)
        else:
            logger.debug(
                f"`to_sql` DEBUG mode, not deleting the files. CSV file is at "
                f"{csv_file_path}, format file is at {fmt_file_path}"
            )



# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 23:07:15 2019

@author: ydima
"""

import logging
import os
from pathlib import Path
import random
import shlex
import string
from subprocess import PIPE, Popen
import tempfile
from typing import Dict, List, Optional, Union

import pandas as pd

from .constants import (
    DIRECTIONS,
    IN,
    IS_WIN32,
    NEWLINE,
    OUT,
    QUERY,
    QUERYOUT,
    SQLCHAR,
    TABLE,
    VIEW,
    BCPandasException,
    BCPandasValueError,
    read_data_settings,
    sql_collation,
)

logger = logging.getLogger(__name__)


def bcp(
    sql_item: str,
    direction: str,
    flat_file: str,
    creds,
    print_output: bool,
    sql_type: str = "table",
    schema: str = "dbo",
    format_file_path: str = None,
    batch_size: int = None,
    col_delimiter: str = None,
    row_terminator: str = None,
    bcp_path: Union[str, Path] = None,
):
    """
    See https://docs.microsoft.com/en-us/sql/tools/bcp-utility
    """
    combos = {TABLE: [IN, OUT], QUERY: [QUERYOUT], VIEW: [IN, OUT]}
    direc = direction.lower()
    # validation
    if direc not in DIRECTIONS:
        raise BCPandasValueError(
            f"Param 'direction' must be one of {DIRECTIONS}, you passed {direc}"
        )
    if direc not in combos[sql_type]:
        raise BCPandasValueError(
            f"Wrong combo of direction and SQL object, you passed {sql_type} and {direc} ."
        )

    # auth
    if creds.with_krb_auth:
        auth = ["-T"]
    else:
        auth = ["-U", creds.username, "-P", creds.password]

    # prepare SQL item string
    if sql_type == QUERY:
        # remove newlines for queries, otherwise messes up BCP
        sql_item_string = quote_this("".join(sql_item.splitlines()))
    else:
        sql_item_string = f"{schema}.{sql_item}"

    # construct BCP command
    bcp_command = [
        "bcp" if bcp_path is None else quote_this(str(bcp_path)),
        sql_item_string,
        direc,
        flat_file,
        "-S",
        creds.server,
        "-d",
        creds.database,
        "-q",  # Executes the SET QUOTED_IDENTIFIERS ON statement, needed for Azure SQL DW
    ] + auth

    if batch_size:
        bcp_command += ["-b", str(batch_size)]

    # formats
    if direc == IN:
        bcp_command += ["-f", format_file_path]
    elif direc in (OUT, QUERYOUT):
        bcp_command += [
            "-c",  # marking as character data, not Unicode (maybe make as param?)
            quote_this(
                f"-t{read_data_settings['delimiter'] if col_delimiter is None else col_delimiter}"
            ),
            quote_this(
                f"-r{read_data_settings['newline'] if row_terminator is None else row_terminator}"
            ),
        ]

    # execute
    bcp_command_log = [c if c != creds.password else "[REDACTED]" for c in bcp_command]
    logger.info(f"Executing BCP command now... \nBCP command is: {bcp_command_log}")
    ret_code = run_cmd(bcp_command, print_output=print_output)
    if ret_code:
        raise BCPandasException(f"Bcp command failed with exit code {ret_code}")


def get_temp_file() -> str:
    """
    Returns full path to a temporary file without creating it.
    """
    tmp_dir = tempfile.gettempdir()
    file_path = os.path.join(
        tmp_dir, "".join(random.choices(string.ascii_letters + string.digits, k=21))
    )
    return file_path


def _escape(input_string: str) -> str:
    """
    Adopted from https://github.com/titan550/bcpy/blob/master/bcpy/format_file_builder.py#L25
    """
    return (
        input_string.replace('"', '\\"')
        .replace("'", "\\'")
        .replace("\r", "\\r")
        .replace("\n", "\\n")
    )


def build_format_file(
    df: pd.DataFrame, delimiter: str, db_cols_order: Optional[Dict[str, int]] = None
) -> str:
    """
    Creates the non-xml SQL format file. Puts 4 spaces between each section.
    See https://docs.microsoft.com/en-us/sql/relational-databases/import-export/non-xml-format-files-sql-server
    for the specification of the file.

    # TODO add params/options to control:
    #   - the char type (not just SQLCHAR),

    Parameters
    ----------
    df : pandas DataFrame
    delimiter : a valid delimiter character
    db_cols_order : dict, optional
        Dict of {database column name -> ordinal position of the column}.
        Maps existing columns in the database to their ordinal position, i.e. the order of the columns in the db table.
        1-indexed, so the first columns is 1, second is 2, etc.
        Only needed if the order of the columns in the dataframe doesn't match the database.

    Returns
    -------
    A string containing the format file
    """
    _space = " " * 4
    format_file_str = f"9.0\n{len(df.columns)}\n"  # Version and Number of columns
    for col_num, col_name in enumerate(df.columns, start=1):
        # last col gets a newline sep
        _delim = delimiter if col_num != len(df.columns) else NEWLINE
        _line = _space.join(
            [
                str(col_num),  # Host file field order
                SQLCHAR,  # Host file data type
                str(0),  # Prefix length
                str(0),  # Host file data length
                f'"{_escape(_delim)}"',  # Terminator (see note below)
                str(
                    col_num if not db_cols_order else db_cols_order[str(col_name)]
                ),  # Server column order
                str(col_name),  # Server column name, optional as long as not blank
                sql_collation,  # Column collation
                "\n",
            ]
        )
        format_file_str += _line
    # FYI very important to surround the Terminator with quotes, otherwise BCP fails with:
    # "Unexpected EOF encountered in BCP data-file". Hugely frustrating bug.
    return format_file_str


def quote_this(this: str, skip: bool = False) -> str:
    """
    OS-safe way to quote a string.

    Returns the string with quotes around it.
    On Windows ~~it's double quotes~~ we skip quoting, 
    on Linux it's single quotes.
    """
    if isinstance(this, str):
        if IS_WIN32:
            return this  # TODO maybe change?
        else:
            return shlex.quote(this)
    else:
        return this


def run_cmd(cmd: List[str], *, print_output: bool) -> int:
    """
    Runs the given command. 
    
    Prints STDOUT in real time,  prints STDERR when command is complete, 
    and logs both STDOUT and STDERR.

    Paramters
    ---------
    cmd : list of str
        The command to run, to be submitted to `subprocess.Popen()`
    print_output: bool
        Whether to print output to STDOUT in real time.
        Regardless, the output will be logged.

    Returns
    -------
    The exit code of the command
    """
    if IS_WIN32:
        with_shell = False
    else:
        with_shell = True
        cmd = " ".join(cmd)  # type: ignore
    proc = Popen(cmd, stdout=PIPE, stderr=PIPE, encoding="utf-8", errors="utf-8", shell=with_shell,)
    # live stream STDOUT
    while True:
        outs = proc.stdout.readline()  # type: ignore[union-attr]
        if outs:
            if print_output:
                print(outs, end="")
            logger.info(outs)
        if proc.poll() is not None and outs == "":
            break
    errs = proc.stderr.readlines()  # type: ignore[union-attr]
    if errs:
        if print_output:
            print(errs, end="")
        logger.error(errs)
    return proc.returncode



from setuptools import find_packages, setup

with open("./README.md", "r") as file:
    long_description = file.read()


setup(
    name="bcpandas",
    version="1.3.0",
    author="yehoshuadimarsky",
    description="High-level wrapper around BCP for high performance data transfers between pandas and SQL Server. No knowledge of BCP required!!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yehoshuadimarsky/bcpandas",
    packages=find_packages(),
    install_requires=["pandas>=0.19", "pyodbc", "sqlalchemy>=1.1.4, <=1.3.17"],
    python_requires=">=3.6, <3.9",
    keywords="bcp mssql pandas",
    entry_points={"pandas.sql.engine": ["bcpandas = bcpandas.main:to_sql"]},
    classifiers=[
        "Topic :: Database",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: SQL",
    ],
)
